from abc import ABC, abstractmethod
from rlgym.utils.gamestates import GameState, PlayerData


class RewardFunction(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def reset(self, optional_data=None):
        raise NotImplementedError

    @abstractmethod
    def get_reward(self, player: PlayerData, state: GameState, optional_data=None):
        raise NotImplementedError

    @abstractmethod
    def get_final_reward(self, player: PlayerData, state: GameState, optional_data=None):
        raise NotImplementedError



import win32api
import win32event
import win32file
import win32pipe
import winerror
import pywintypes
from enum import Enum
import numpy as np
import sys


class State(Enum):
    IDLE = 0
    CONNECTING = 1
    SETTING_UP = 2
    RECEIVING = 3
    SENDING = 4
    AWAITING_ACTION = 5
    DIED = 6


class DistributedGym:
    def __init__(self, matches: list, wait_count=sys.maxsize, wait_ratio=1.0):
        self._matches = matches
        self._next_match = 0
        self.wait_count = wait_count
        self.wait_ratio = wait_ratio

        self._PIPE_SIZE = 4096
        self._PIPE_NAME = r'\\.\pipe\RLGym'

        self.events = []
        self.runners = []
        self.runners_by_id = {}
        self.next_id = 0

        self.notify_reset = []
        self.ids = []
        self.obs = []
        self.rewards = []
        self.t_ids = []
        self.t_obs = []
        self.t_rewards = []

        # All envs must have the same dimensions
        first_env = matches[0]
        self.observation_space = first_env.observation_space
        self.action_space = first_env.action_space

        self.idle_runners = 0
        self.waiting_runners = 0
        self._create_runner()

    def _create_runner(self):
        env = self._get_next_env()
        if env.observation_space != self.observation_space:
            raise RuntimeError('Invalid env obs space. Expected:', self.observation_space, 'Got:', env.observation_space)
        if env.action_space != self.action_space:
            raise RuntimeError('Invalid env action space. Expected:', self.action_space, 'Got:', env.action_space)

        print("Building win32 pipe...")
        event = win32event.CreateEvent(None, True, True, None)
        pipe = win32pipe.CreateNamedPipe(self._PIPE_NAME, win32pipe.PIPE_ACCESS_DUPLEX | win32file.FILE_FLAG_OVERLAPPED,
                                         win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_READMODE_MESSAGE | win32pipe.PIPE_WAIT,
                                         win32pipe.PIPE_UNLIMITED_INSTANCES, self._PIPE_SIZE, self._PIPE_SIZE, 0, None)
        overlap = pywintypes.OVERLAPPED()
        overlap.hEvent = event
        r_id = self._get_new_id()
        print("Pipe built, initializing runner...")
        runner = Runner(r_id, env, pipe, overlap, self._PIPE_SIZE)

        self.events.append(event)
        self.runners.append(runner)
        self.runners_by_id[r_id] = runner
        self.idle_runners += 1

    def _get_next_env(self):
        env_name = self._envs[self._next_env]
        self._next_env += 1
        if self._next_env == len(self._envs):
            self._next_env = 0
        return self._get_env(env_name)

    def _get_new_id(self):
        r_id = self.next_id
        self.next_id += 1
        return r_id

    def reset(self):
        print("Resetting distributed environment...")
        # Notify reset to env objects
        for r_id in self.notify_reset:
            self.runners_by_id[r_id].env.episode_reset()
        self.notify_reset.clear()
        # Wait for N observations
        ids, obs, rewards, done = self._get_next_batch(is_reset=True)
        assert not done
        return ids, obs, rewards

    def step(self, ids, actions):
        # Add actions to runners
        for i in range(len(ids)):
            r_id, agent = ids[i]
            action = actions[i]
            runner = self.runners_by_id[r_id]
            if runner:
                runner.add_action(agent, action)

        # Send actions of ready runners
        for r in self.runners:
            if r.send_actions():
                self.waiting_runners -= 1

        # Get N obs if not terminal, only t_obs otherwise
        ids, obs, rewards, done = self._get_next_batch()

        return ids, obs, rewards, done

    def close(self):
        for runner in self.runners:
            runner.close()
        for event in self.events:
            win32file.CloseHandle(event)

    def _get_next_batch(self, is_reset=False):
        # If done send only terminal obs and get next full batch on "reset"
        while self.waiting_runners < min(self.wait_count, len(self.runners) - self.idle_runners) \
                or self.waiting_runners < int(self.wait_ratio * (len(self.runners) - self.idle_runners)) \
                or len(self.runners) - self.idle_runners == 0:
            # Wait for signal
            code = win32event.WaitForMultipleObjects(self.events, False, win32event.INFINITE)
            index = code - win32event.WAIT_OBJECT_0

            # Check index out of bounds? Server crashed anyways
            runner = self.runners[index]
            if runner.pending_op:
                try:
                    result = win32file.GetOverlappedResult(runner.pipe, runner.overlap, 0)

                    if runner.state == State.CONNECTING:
                        # if result == 0: # GetOverlappedResult throws an exception in python
                        #     raise RuntimeError('Connection failed', win32api.GetLastError())
                        runner.state = State.SETTING_UP
                    elif runner.state == State.SENDING:
                        if result == 0:
                            runner.state = State.DIED
                        else:
                            runner.state = State.RECEIVING
                    elif runner.state == State.RECEIVING:
                        if result == 0:
                            runner.state = State.DIED
                        else:
                            runner.b_read = result
                            runner.state = State.SENDING
                    else:
                        print('Invalid runner state', runner.state, runner.pending_op)
                        runner.state = State.DIED
                except:
                    print('GetOverlappedResult failed. Error code:', win32api.GetLastError(), 'Resetting runner', runner.id)
                    runner.state = State.DIED

                runner.pending_op = False

            if runner.state == State.SETTING_UP:
                # Send setup, state changes internally to sending
                print('Setting up runner ', runner.id)
                runner.setup()
                self.idle_runners -= 1
                if self.idle_runners == 0:
                    self._create_runner()
            elif runner.state == State.SENDING:
                # get state from buffer and build obs
                obs, rewards, terminal = runner.get_obs()
                # if terminal then read first state else wait action
                if terminal:
                    self.t_ids.extend([(runner.id, agent) for agent in range(len(obs))])
                    self.t_obs.extend(obs)
                    self.t_rewards.extend(rewards)
                    self.notify_reset.append(runner.id)
                    runner.state = State.RECEIVING
                else:
                    self.ids.extend([(runner.id, agent) for agent in range(len(obs))])
                    self.obs.extend(obs)
                    self.rewards.extend(rewards)
                    # can't send action instantly, have to wait for action to be ready.
                    runner.state = State.AWAITING_ACTION
                    self.waiting_runners += 1
                    # unset event manually?
                    win32event.ResetEvent(self.events[index])
            elif runner.state == State.RECEIVING:
                # queue read operation on pipe. All contained in receive_state method
                runner.receive_state()
            elif runner.state == State.DIED:
                print('Runner', runner.id, 'died.')
                self.idle_runners += 1
                if runner.id in self.notify_reset:
                    self.notify_reset.remove(runner.id)
                new_id = self._get_new_id()
                self.runners_by_id[runner.id] = None
                self.runners_by_id[new_id] = runner
                runner.reset(new_id)
            else:
                print('Invalid runner state', runner.state, runner.pending_op)
                runner.state = State.DIED

        if not is_reset and len(self.notify_reset) > 0:
            terminal = True
            ids = self.t_ids
            obs = self.t_obs
            rewards = self.t_rewards
            self.t_ids = []
            self.t_obs = []
            self.t_rewards = []
        else:
            terminal = False
            ids = self.ids
            obs = self.obs
            rewards = self.rewards
            self.ids = []
            self.obs = []
            self.rewards = []

        return ids, obs, rewards, terminal


class Runner:
    def __init__(self, r_id, env, pipe, overlap, buffer_size):
        self.id = r_id
        self.env = env
        self.pipe = pipe
        self.overlap = overlap
        self.state = State.IDLE
        self.pending_op = False
        self.b_read = 0
        self.r_buffer = win32file.AllocateReadBuffer(buffer_size)
        self.w_buffer = None
        self.action_count = 0
        self.actions = np.empty((env.agents, env.action_space), dtype=float)

        self.connect()

    def connect(self):
        print("Connecting pipe to DLL...")
        code = win32pipe.ConnectNamedPipe(self.pipe, self.overlap)
        if code == winerror.ERROR_IO_PENDING:
            self.state = State.CONNECTING
            self.pending_op = True
            print("Unable to connect...")
        else:
            self.state = State.SETTING_UP
            # is this necessary? It was like this on msoft's docs
            win32event.SetEvent(self.overlap.hEvent)
            print("Pipe connected!")

    def setup(self):
        self.w_buffer = str.encode(self.env.get_config())
        self._send_data()

    def _send_data(self):
        try:
            code, _ = win32file.WriteFile(self.pipe, self.w_buffer, self.overlap)
            # signal set automatically after write?
            if code == 0:
                self.state = State.RECEIVING
            else:
                self.state = State.SENDING
                self.pending_op = True
        except:
            print('WriteFile failed. Error code:', win32api.GetLastError(), 'Resetting runner', self.id)
            self.state = State.DIED
            win32event.SetEvent(self.overlap.hEvent)

    def receive_state(self):
        try:
            print("Receiving state...")
            code, _ = win32file.ReadFile(self.pipe, self.r_buffer, self.overlap)
            # signaled if read returns instantly?
            # if code == 0: Pywin expects you to get readable bytes from GetOverlappedResult
            #     self.state = State.SENDING
            if code == 0 or code == winerror.ERROR_IO_PENDING:
                self.state = State.RECEIVING
                self.pending_op = True
                print("State received!")
            else:
                print('ReadFile failed. Error code:', win32api.GetLastError(), 'Resetting runner', self.id)
                self.state = State.DIED
                win32event.SetEvent(self.overlap.hEvent)
        except:
            print('ReadFile failed. Error code:', win32api.GetLastError(), 'Resetting runner', self.id)
            self.state = State.DIED
            win32event.SetEvent(self.overlap.hEvent)

    def get_obs(self):
        state = self.env.parse_state(bytes(self.r_buffer[:self.b_read]).decode())
        obs = self.env.build_observations(state)
        rewards = self.env.get_rewards(state)
        done = self.env.is_done(state)
        return obs, rewards, done

    def add_action(self, agent, action):
        self.actions[agent] = action
        self.action_count += 1

    def send_actions(self):
        if self.state == State.AWAITING_ACTION and self.action_count == self.env.agents:
            self.w_buffer = str.encode(self.env.format_actions(self.actions))
            self._clear_actions()
            self._send_data()
            return True
        return False

    def _clear_actions(self):
        self.action_count = 0

    def reset(self, new_id):
        win32pipe.DisconnectNamedPipe(self.pipe)
        self.id = new_id
        self._clear_actions()
        self.connect()

    def close(self):
        win32file.CloseHandle(self.pipe)


from rlgym.utils.terminal_conditions import TerminalCondition

class TimeoutCondition(TerminalCondition):
    def __init__(self, max_ticks):
        super().__init__()
        self.ticks = 0
        self.max_ticks = max_ticks

    def reset(self, optional_data=None):
        self.ticks = 0

    def is_terminal(self, state, optional_data=None):
        self.ticks += 1
        return self.ticks >= self.max_ticks

    def look_ahead(self, state, optional_data=None):
        return self.ticks + 1 >= self.max_ticks


class GoalScoredCondition(TerminalCondition):
    def __init__(self):
        super().__init__()
        self.blue_score = 0
        self.orange_score = 0

    def reset(self, optional_data=None):
        pass

    def is_terminal(self, state, optional_data=None):
        if state.blue_score != self.blue_score or state.orange_score != self.orange_score:
            self.blue_score = state.blue_score
            self.orange_score = state.orange_score
            return True
        return False

    def look_ahead(self, state, optional_data=None):
        if state.blue_score != self.blue_score or state.orange_score != self.orange_score:
            return True
        return False


from abc import abstractmethod


class Environment:
    def __init__(self):
        self.observation_space = None
        self.action_space = None
        self.agents = None
        self.bots = None
        self._team_size = None

    @abstractmethod
    def get_config(self):
        raise NotImplementedError

    @abstractmethod
    def episode_reset(self):
        raise NotImplementedError

    @abstractmethod
    def build_observations(self, state):
        raise NotImplementedError

    @abstractmethod
    def get_rewards(self, next_state):
        raise NotImplementedError

    @abstractmethod
    def is_done(self, state):
        raise NotImplementedError

    @abstractmethod
    def parse_state(self, state_string):
        raise NotImplementedError

    @abstractmethod
    def format_actions(self, actions):
        raise NotImplementedError




import os
from typing import List


def make(env_name: str,
         ep_len_minutes: float = 5,
         game_speed: int = 100,
         tick_skip: int = 8,
         spawn_opponents: bool = True,
         random_resets: bool = False,
         team_size: int = None,
         terminal_conditions: List[object] = None,
         reward_fn: object = None,
         obs_builder: object = None,
         path_to_rl: str = None,
         use_injector: bool = False):
    """
    :param env_name: Name of your env, can be any of (Custom, Duel, Doubles, Standard) with or without Self
    :param ep_len_minutes: The episode length in minutes, seconds granularity can be achieved with a float
    :param game_speed: The speed the physics will run at, leave it at 100 unless your game can't run at over 240fps
    :param tick_skip: The amount of physics ticks your action will be repeated for
    :param spawn_opponents: Whether you want opponents or not
    :param random_resets: If enabled cars and ball will spawn in random locations after every reset
    :param team_size: Cars per team
    :param terminal_conditions: List of terminal condition objects (rlgym.utils.TerminalCondition)
    :param reward_fn: Reward function object (rlgym.utils.RewardFunction)
    :param obs_builder: Observation builder object (rlgym.utils.ObsBuilder)
    :param path_to_rl: Path to RocketLeague executable, this is optional
    :param use_injector: Whether to use a custom injector or not
    :return: Gym object
    """

    # Imports are inside the function because setup fails otherwise (Missing win32file)
    from rlgym.gym import Gym
    from rlgym.envs import match_factory
    from rlgym.version import print_current_release_notes

    print_current_release_notes()

    custom_args = dict(ep_len_minutes=ep_len_minutes, game_speed=game_speed, tick_skip=tick_skip,
                       spawn_opponents=spawn_opponents, random_resets=random_resets, team_size=team_size,
                       terminal_conditions=terminal_conditions, reward_fn=reward_fn, obs_builder=obs_builder)
    match = match_factory.build_match(env_name, **custom_args)
    if match is None:
        raise ValueError("RLGym was unable to construct match!\n"
                         "Match ID: {}\n"
                         "Custom match params: {}".format(env_name, custom_args))

    return Gym(match, pipe_id=os.getpid(), path_to_rl=path_to_rl, use_injector=use_injector)


# FIXME i think this doesn't work anymore [env = Gym]
# For a distributed architecture we recommend using many normal Gyms instead
# Deprecated
def make_distributed(env_names, custom_arg_dicts=None, path_to_rl=None, use_injector: bool = False):
    from rlgym.gym import Gym
    from rlgym.distributed_gym import DistributedGym
    from rlgym.envs import match_factory
    from rlgym.version import print_current_release_notes

    print_current_release_notes()

    envs = []
    for i in range(len(env_names)):
        name = env_names[i]
        args = custom_arg_dicts[i]
        match = match_factory.build_match(name, custom_args=args)
        if match is not None:
            env = Gym(match, pipe_id=i)
            envs.append(env)
        else:
            RuntimeError("RLGym was unable to construct match while building matches for distributed env!"
                         "\nMatch ID: {}"
                         "\nCustom match params: {}".format(name, args))

    if len(envs) == 0:
        raise RuntimeError("RLGym was unable to build any matches for distributed env!")

    return DistributedGym(envs)



from rlgym.utils import math
import numpy as np
from typing import Optional


class PhysicsObject(object):
    def __init__(self, position=None, quaternion=None, linear_velocity=None, angular_velocity=None):
        self.position: np.ndarray = position if position else np.zeros(3)
        self.quaternion: np.ndarray = quaternion if quaternion else np.zeros(4)
        self.linear_velocity: np.ndarray = linear_velocity if linear_velocity else np.zeros(3)
        self.angular_velocity: np.ndarray = angular_velocity if angular_velocity else np.zeros(3)
        self._euler_angles: Optional[np.ndarray] = None
        self._rotation_mtx: Optional[np.ndarray] = None

    def decode_car_data(self, car_data: np.ndarray):
        self.position = car_data[:3]
        self.quaternion = car_data[3:7]
        self.linear_velocity = car_data[7:10]
        self.angular_velocity = car_data[10:]

    # roll, pitch, yaw
    def euler_angles(self) -> np.ndarray:
        if self._euler_angles is None:
            self._euler_angles = math.quat_to_euler(self.quaternion)
        return self._euler_angles

    def rotation_mtx(self) -> np.ndarray:
        if self._rotation_mtx is None:
            self._rotation_mtx = math.quat_to_rot_mtx(self.quaternion)
        return self._rotation_mtx

    def forward(self) -> np.ndarray:
        return self.rotation_mtx()[:, 0]

    def right(self) -> np.ndarray:
        return self.rotation_mtx()[:, 1]

    def up(self) -> np.ndarray:
        return self.rotation_mtx()[:, 2]

    def decode_ball_data(self, ball_data: np.ndarray):
        self.position = ball_data[:3]
        self.linear_velocity = ball_data[3:6]
        self.angular_velocity = ball_data[6:9]

    def serialize(self):
        repr = []

        if self.position is not None:
            for arg in self.position:
                repr.append(arg)
                
        if self.quaternion is not None:
            for arg in self.quaternion:
                repr.append(arg)
                
        if self.linear_velocity is not None:
            for arg in self.linear_velocity:
                repr.append(arg)
                
        if self.angular_velocity is not None:
            for arg in self.angular_velocity:
                repr.append(arg)

        if self._euler_angles is not None:
            for arg in self._euler_angles:
                repr.append(arg)

        return repr



from typing import List, Union, Tuple

import numpy as np
from gym import Env

from rlgym.utils import BotRecorder
from rlgym.gamelaunch import launch_rocket_league
from rlgym.communication import CommunicationHandler, Message
from rlgym.utils.gamestates import GameState


class Gym(Env):
    """
    The Rocket League gym environment.
    """
    def __init__(self, match, pipe_id=0, path_to_rl=None, use_injector=False):
        super().__init__()

        self._match = match
        self.observation_space = match.observation_space
        self.action_space = match.action_space

        self._path_to_rl = path_to_rl
        self._use_injector = use_injector

        self._comm_handler = CommunicationHandler()
        self._local_pipe_name = self._comm_handler.format_pipe_id(pipe_id)
        self._local_pipe_id = pipe_id

        self._game_process = None

        self._open_game()
        self._setup_plugin_connection()

        self._recorder = BotRecorder(self._comm_handler)
        self._prev_state = None

    def _open_game(self):
        print("Launching Rocket League, make sure bakkesmod is running.")
        # Game process is only set if launched with path_to_rl
        self._game_process = launch_rocket_league(self._local_pipe_name, self._path_to_rl, self._local_pipe_name)

    def _setup_plugin_connection(self):
        self._comm_handler.open_pipe(self._local_pipe_name)
        self._comm_handler.send_message(header=Message.RLGYM_CONFIG_MESSAGE_HEADER, body=self._match.get_config())

    def reset(self) -> np.ndarray:
        exception = self._comm_handler.send_message(header=Message.RLGYM_RESET_GAME_STATE_MESSAGE_HEADER, body=Message.RLGYM_NULL_MESSAGE_BODY)
        if exception is not None:
            self._attempt_recovery()
            exception = self._comm_handler.send_message(header=Message.RLGYM_RESET_GAME_STATE_MESSAGE_HEADER,
                                                        body=Message.RLGYM_NULL_MESSAGE_BODY)
            if exception is not None:
                import sys
                print("!UNABLE TO RECOVER ROCKET LEAGUE!\nEXITING")
                sys.exit(-1)

        # print("Sending reset command")
        self._match.episode_reset()
        state = self._receive_state()
        self._prev_state = state

        #self.recorder.reset()

        return self._match.build_observations(state)

    def step(self, actions: Union[np.ndarray, List[np.ndarray], List[float]]) -> Tuple[List, List, bool, GameState]:
        # print("Stepping")
        #self._parse_tanh_actions(actions)
        actions_sent = self._send_actions(actions)
        # print("Requesting state")

        received_state = self._receive_state()
        if received_state is None:  # TODO ask matt why?
            state = self._prev_state
        else:
            state = received_state

        #self.recorder.step(state)
        # print("Building obs")
        obs = self._match.build_observations(state)
        # print("Getting rewards")
        reward = self._match.get_rewards(state)
        # print("Checking done")
        done = self._match.is_done(state) or received_state is None or not actions_sent
        self._prev_state = state

        return obs, reward, done, state

    def close(self):
        self._comm_handler.close_pipe()
        if self._game_process is not None:
            self._game_process.terminate()

    def _receive_state(self):
        # print("Waiting for state...")
        message, exception = self._comm_handler.receive_message(header=Message.RLGYM_STATE_MESSAGE_HEADER)
        if exception is not None:
            self._attempt_recovery()
            return None

        if message is None:
            return None
        # print("GOT MESSAGE\n HEADER: '{}'\nBODY: '{}'\n".format(message.header, message.body))
        if message.body is None:
            return None

        return self._match.parse_state(message.body)

    def _send_actions(self, actions):
        action_string = self._match.format_actions(actions)
        #print("Transmitting actions",action_string,"...")
        exception = self._comm_handler.send_message(header=Message.RLGYM_AGENT_ACTION_IMMEDIATE_RESPONSE_MESSAGE_HEADER, body=action_string)
        if exception is not None:
            self._attempt_recovery()
            return False
        return True
        # print("Message sent", action_string, "...")

    def seed(self, seed):
        #TODO: ensure that nothing actually needs to be seeded. I don't think any rng are used here, but need to make sure.
        pass

    def _parse_tanh_actions(self, actions):
        choices = [1, 0]

        prob = (1.0 + actions[-1]) / 2.0
        if prob is None or np.isnan(prob):
            prob = 0
        prob = min(max(prob, 0), 1)
        choice = np.random.choice(choices, p=[prob, 1 - prob])
        actions[-1] = choice

        prob = (1.0 + actions[-2]) / 2.0
        if prob is None or np.isnan(prob):
            prob = 0
        prob = min(max(prob, 0), 1)
        choice = np.random.choice(choices, p=[prob, 1 - prob])
        actions[-2] = choice

        prob = (1.0 + actions[-3]) / 2.0
        if prob is None or np.isnan(prob):
            prob = 0
        prob = min(max(prob, 0), 1)
        choice = np.random.choice(choices, p=[prob, 1 - prob])
        actions[-3] = choice

    def _attempt_recovery(self):
        print("!ROCKET LEAGUE HAS CRASHED!\nATTEMPTING RECOVERY")
        import os
        import time
        self.close()
        proc_list = os.popen('wmic process get description, processid').read()
        num_instances = proc_list.count("RocketLeague.exe")
        wait_time = 2 * num_instances

        print("Discovered {} existing Rocket League processes. Waiting {} seconds before attempting to open "
              "a new one.".format(num_instances, wait_time))

        time.sleep(wait_time)
        self._open_game()
        self._setup_plugin_connection()



from rlgym.envs import Match
from rlgym.utils import common_values
from rlgym.utils.terminal_conditions import common_conditions
from rlgym.utils.reward_functions import ShootBallReward
from rlgym.utils.obs_builders import RhobotObs


def basic_duel_match(**kwargs):
    if "team_size" not in kwargs.keys() or kwargs["team_size"] is None:
        kwargs["team_size"] = 1

    game_speed, tick_skip, spawn_opponents, random_resets, self_play, team_size, terminal_conditions, reward_fn, obs_builder \
        = get_default_params(**kwargs)

    return Match(team_size=team_size,
                 tick_skip=tick_skip,
                 game_speed=game_speed,
                 spawn_opponents=spawn_opponents,
                 random_resets=random_resets,
                 self_play=self_play,
                 reward_function=reward_fn,
                 terminal_conditions=terminal_conditions,
                 obs_builder=obs_builder)


def basic_doubles_match(**kwargs):
    if "team_size" not in kwargs.keys() or kwargs["team_size"] is None:
        kwargs["team_size"] = 2

    game_speed, tick_skip, spawn_opponents, random_resets, self_play, team_size, terminal_conditions, reward_fn, obs_builder \
        = get_default_params(**kwargs)

    return Match(team_size=team_size,
                 tick_skip=tick_skip,
                 game_speed=game_speed,
                 spawn_opponents=spawn_opponents,
                 random_resets=random_resets,
                 self_play=self_play,
                 reward_function=reward_fn,
                 terminal_conditions=terminal_conditions,
                 obs_builder=obs_builder)


def basic_standard_match(**kwargs):
    if "team_size" not in kwargs.keys() or kwargs["team_size"] is None:
        kwargs["team_size"] = 3

    game_speed, tick_skip, spawn_opponents, random_resets, self_play, team_size, terminal_conditions, reward_fn, obs_builder \
        = get_default_params(**kwargs)

    return Match(team_size=team_size,
                 tick_skip=tick_skip,
                 game_speed=game_speed,
                 spawn_opponents=spawn_opponents,
                 random_resets=random_resets,
                 self_play=self_play,
                 reward_function=reward_fn,
                 terminal_conditions=terminal_conditions,
                 obs_builder=obs_builder)


def get_default_params(**kwargs):
    self_play = False
    if "self_play" in kwargs.keys():
        self_play = kwargs["self_play"]

    ep_len_minutes = 45 / 60
    if "ep_len_minutes" in kwargs.keys():
        ep_len_minutes = kwargs["ep_len_minutes"]

    game_speed = 100
    if "game_speed" in kwargs.keys():
        game_speed = kwargs["game_speed"]

    tick_skip = 8
    if "tick_skip" in kwargs.keys():
        tick_skip = kwargs["tick_skip"]

    ticks_per_sec = 120
    ticks_per_min = ticks_per_sec * 60
    max_ticks = int(round(ep_len_minutes * ticks_per_min / tick_skip))

    spawn_opponents = True
    if "spawn_opponents" in kwargs.keys():
        spawn_opponents = kwargs["spawn_opponents"]

    random_resets = False
    if "random_resets" in kwargs.keys():
        random_resets = kwargs["random_resets"]

    team_size = 1
    if "team_size" in kwargs.keys():
        team_size = kwargs["team_size"]

    terminal_conditions = [common_conditions.TimeoutCondition(max_ticks), common_conditions.GoalScoredCondition()]
    if "terminal_conditions" in kwargs.keys() and kwargs["terminal_conditions"] is not None:
        terminal_conditions = kwargs["terminal_conditions"]

    reward_fn = ShootBallReward()
    if "reward_fn" in kwargs.keys() and kwargs["reward_fn"] is not None:
        reward_fn = kwargs["reward_fn"]

    obs_builder = RhobotObs()
    if "obs_builder" in kwargs.keys() and kwargs["obs_builder"] is not None:
        obs_builder = kwargs["obs_builder"]

    return game_speed, tick_skip, spawn_opponents, random_resets, self_play, team_size, terminal_conditions, reward_fn, obs_builder



from abc import ABC, abstractmethod
from rlgym.utils.gamestates import GameState


class TerminalCondition(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def reset(self, optional_data=None):
        raise NotImplementedError

    @abstractmethod
    def is_terminal(self, state: GameState, optional_data=None) -> bool:
        raise NotImplementedError

    @abstractmethod
    def look_ahead(self, state: GameState, optional_data=None) -> bool:
        raise NotImplementedError



import numpy as np
from rlgym.utils import math, common_values
from rlgym.utils.reward_functions import RewardFunction


class ShootBallReward(RewardFunction):
    GOAL_REWARD = 60
    GOAL_PUNISHMENT = -40
    PLAYER_TO_BALL_VEL_WEIGHT = 0.04
    BALL_TO_GOAL_VEL_WEIGHT = 0.065

    def __init__(self):
        super().__init__()
        self.orange_score = 0
        self.blue_score = 0
        self.last_touch = None

    def reset(self, optional_data=None):
        self.last_touch = None

    def get_reward(self, player, state, optional_data=None):
        self.last_touch = state.last_touch

        b_rew = self._get_player_ball_reward(player, state) * ShootBallReward.PLAYER_TO_BALL_VEL_WEIGHT
        g_rew = self._get_ball_goal_reward(player, state) * ShootBallReward.BALL_TO_GOAL_VEL_WEIGHT
        #print(g_rew)

        #print("{}  |  {:3.6f}  |  {:3.6f}".format(player.team_num, b_rew, g_rew))
        return b_rew + g_rew

    def get_final_reward(self, player, state, optional_data=None):
        return self._get_goal_reward(player, state)

    def _get_ball_goal_reward(self, player, state):
        if player.team_num == common_values.BLUE_TEAM:
            ball = state.ball
        else:
            ball = state.inv_ball

        b_vel = ball.linear_velocity
        b_pos = ball.position
        g_pos = common_values.ORANGE_GOAL_CENTER

        dist = math.get_dist(g_pos, b_pos)
        vel_to_goal = math.scalar_projection(b_vel, dist)

        return vel_to_goal / 100

    def _get_player_ball_reward(self, player, state):
        if player.team_num == common_values.BLUE_TEAM:
            ball = state.ball
            car = player.car_data
        else:
            ball = state.inv_ball
            car = player.inverted_car_data

        p_vel = car.linear_velocity
        b_pos = ball.position
        p_pos = car.position

        dist = math.get_dist(b_pos, p_pos)
        vel_to_ball = math.scalar_projection(p_vel, dist)

        return vel_to_ball / 100

    def _get_goal_reward(self, player, state):
        os = state.orange_score
        bs = state.blue_score
        team = player.team_num

        if os != self.orange_score:
            self.orange_score = os
            if team == common_values.ORANGE_TEAM and self.last_touch == player.car_id:
                return ShootBallReward.GOAL_REWARD
            return ShootBallReward.GOAL_PUNISHMENT

        if bs != self.blue_score:
            self.blue_score = bs
            if team == common_values.BLUE_TEAM and self.last_touch == player.car_id:
                return ShootBallReward.GOAL_REWARD
            return ShootBallReward.GOAL_PUNISHMENT

        return 0



# From https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/version.py
# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

__version__ = '0.2.0'

release_notes = {
    '0.2.0': """
    - Switched from custom_args dict to keyword args in rlgym.make()
    """,
    '0.1.4': """
    - Gym now inherits openai.gym.Env - Chainso
    - Fixed bug where rlgym crashed trying to parse the state string in some scenarios
    """,
    '0.1.3': """
    - Renamed PhysicsObject.orientation to euler_angles
    - euler_angles is now a function so they aren't calculated if not needed
    - Added rotation_mtx, forward, right and up vectors to PhysicsObject (only usable for the car)
    - Removed conversion to np array for rewards and observations
    - Fixed bug in RhobotObs
    """,
    '0.1.2': """
    - Obs are now numpy arrays
    - Added types to core classes to make customization easier
    - Fixed bakkesmod bug
    """,
    '0.1.1': """
    Added missing scipy dependency
    """,
    '0.1.0': """
    Initial Release
    """
}


def get_current_release_notes():
    if __version__ in release_notes:
        return release_notes[__version__]
    return ''


def print_current_release_notes():
    print(f"Version {__version__}")
    print(get_current_release_notes())
    print("")




class Message(object):
    RLGYM_HEADER_END_TOKEN = "RLGEHEADER"
    RLGYM_BODY_END_TOKEN = "RLGEBODY"

    RLGYM_NULL_MESSAGE_HEADER = "RLGNMH"
    RLGYM_NULL_MESSAGE_BODY = "RLGNMB"
    RLGYM_CONFIG_MESSAGE_HEADER = "RLGC"

    RLGYM_STATE_MESSAGE_HEADER = "RLGSMH"
    RLGYM_AGENT_ACTION_MESSAGE_HEADER = "RLGAAMH"
    RLGYM_RESET_GAME_STATE_MESSAGE_HEADER = "RLGRGSMH"
    RLGYM_AGENT_ACTION_IMMEDIATE_RESPONSE_MESSAGE_HEADER = "RLGAAIRMH"

    RLGYM_REQUEST_LAST_BOT_INPUT_MESSAGE_HEADER = "RLGRLBIMH"
    RLGYM_LAST_BOT_INPUT_MESSAGE_HEADER = "RLGLBIMH"


    def __init__(self, header=None, body=None):
        if header is None:
            header = Message.RLGYM_NULL_MESSAGE_HEADER
        if body is None:
            body = Message.RLGYM_NULL_MESSAGE_BODY

        self.body = body
        self.header = header

    def serialize(self):
        return "{header}{header_token}{body}{body_token}\0".format(header=self.header,
                                                                   header_token=Message.RLGYM_HEADER_END_TOKEN,
                                                                   body=self.body,
                                                                   body_token=Message.RLGYM_BODY_END_TOKEN)

    def deserialize(self, serialized_str):
        s = serialized_str

        header = s[:s.find(Message.RLGYM_HEADER_END_TOKEN)]
        start = s.find(Message.RLGYM_HEADER_END_TOKEN) + len(Message.RLGYM_HEADER_END_TOKEN)
        end = s.find(Message.RLGYM_BODY_END_TOKEN)
        body = s[start:end]

        self.body = body
        self.header = header


from rlgym.communication import Message, CommunicationHandler
from rlgym.utils import math, common_values
from rlgym.utils.obs_builders import ObsBuilder
import numpy as np

class BotRecorder(object):
    def __init__(self, comm_handler):
        self.comm_handler = comm_handler
        self.orange_score = 0
        self.bot_actions = []
        self.bot_obs = []
        self.prev_action = [0 for _ in range(8)]

        self.last_touch = None
        self.player = None

    def step(self, state):
        obs = self.build_obs_for_player(state)
        if len(obs) != 62:
            print("INVALID OBS ENCOUNTERED BY BOT RECORDER:",obs)
            return

        self.request()
        self.bot_obs.append(obs)
        if state.orange_score != self.orange_score and state.last_touch == self.player.car_id:
            self.orange_score = state.orange_score
            self.save()
            self.reset()

    def reset(self):
        self.bot_obs = []
        self.bot_actions = []
        self.prev_action = [0 for _ in range(8)]

    def request(self):
        self.comm_handler.send_message(header=Message.RLGYM_REQUEST_LAST_BOT_INPUT_MESSAGE_HEADER)
        response = self.comm_handler.receive_message(header=Message.RLGYM_LAST_BOT_INPUT_MESSAGE_HEADER)
        self.bot_actions.append(response.body)
        self.prev_action = [float(x.strip()) for x in response.body.split(" ")]

    def save(self):
        print("Bot Recorder Saving", len(self.bot_actions))
        with open("data/recordings/val_recording.txt", 'a') as f:
            for state, action in zip(self.bot_obs, self.bot_actions):
                s = ''.join(["{} ".format(x) for x in state])
                #a = ["{} ".format(x) for x in action]
                line = "{}|{}\n".format(s[:-1], action)#a[:-1])
                f.write(line)
            f.write("END EPISODE\n")

    def build_obs_for_player(self, state):
        if self.player is None:
            for p in state.players:
                if p.team_num == common_values.ORANGE_TEAM:
                    self.player = p
                    break
        player = self.player
        if player is None:
            return []

        prev_actions = self.prev_action
        if prev_actions is None:
            print("ATTEMPTED TO BUILD RHOBOT OBS WITH NO PREV ACTIONS ARGUMENT!")
            raise ArithmeticError

        players = state.players
        if player.team_num == common_values.ORANGE_TEAM:
            player_car = player.inverted_car_data
            ball = state.inv_ball
        else:
            player_car = player.car_data
            ball = state.ball

        ob = [arg for arg in prev_actions]
        ob.append(int(player.has_flip))
        ob.append(int(player.boost_amount))
        ob.append(int(player.on_ground))

        ob += player_car.position
        ob += player_car.euler_angles()
        ob += [np.sin(arg) for arg in player_car.euler_angles()]
        ob += [np.cos(arg) for arg in player_car.euler_angles()]
        ob += player_car.linear_velocity
        ob += player_car.angular_velocity

        ob += ball.position
        ob += ball.linear_velocity
        ob += ball.angular_velocity

        ob += common_values.ORANGE_GOAL_CENTER
        ob += common_values.BLUE_GOAL_CENTER

        for other in players:
            if other.car_id == player.car_id:
                continue

            if other.team_num == common_values.BLUE_TEAM and player.team_num == other.team_num:
                car_data = other.car_data
            else:
                car_data = other.inverted_car_data

            # TODO: COMMENT THIS OUT
            #car_data = ObsBuilder.get_random_physics_state()

            ob += car_data.position
            ob += car_data.euler_angles()
            ob += [np.sin(arg) for arg in car_data.euler_angles()]
            ob += [np.cos(arg) for arg in car_data.euler_angles()]
            ob += car_data.linear_velocity
            ob += car_data.angular_velocity

        return ob


import numpy as np

def get_dist(x,y):
    return np.subtract(x,y)

def vector_projection(vec, dest_vec, mag_squared=None):
    """Compute the vector projection of vec on to dest_vec."""
    if mag_squared is None:
        norm = vecmag(dest_vec)
        if norm == 0:
            return dest_vec
        mag_squared = norm * norm

    if mag_squared == 0:
        return dest_vec

    dot = np.dot(vec, dest_vec)
    projection = np.multiply(np.divide(dot, mag_squared), dest_vec)
    return projection

def scalar_projection(vec, dest_vec):
    """Compute the scalar projection of vec on to dest_vec."""
    norm = vecmag(dest_vec)

    if norm == 0:
        return 0

    dot = np.dot(vec, dest_vec) / norm
    return dot

def squared_vecmag(vec):
    x = np.linalg.norm(vec)
    return x*x

def vecmag(vec):
    norm = np.linalg.norm(vec)
    return norm

def unitvec(vec):
    return np.divide(vec, vecmag(vec))

def quat_to_euler(quat):
    w, x, y, z = quat
    sinr_cosp = 2 * (w * x + y * z)
    cosr_cosp = 1 - 2 * (x*x + y*y)
    sinp = 2 * (w * y - z * x)
    siny_cosp = 2 * (w * z + x * y)
    cosy_cosp = 1 - 2 * (y * y + z * z)

    roll = np.arctan2(sinr_cosp, cosr_cosp)
    if abs(sinp) > 1:
        pitch = np.pi/2
    else:
        pitch = np.arcsin(sinp)
    yaw = np.arctan2(siny_cosp, cosy_cosp)

    return np.array([roll, pitch, yaw])

# From RLUtilities
def quat_to_rot_mtx(quat: np.ndarray) -> np.ndarray:
    s = 1.0 / np.dot(quat, quat)
    w = -quat[0]
    x = -quat[1]
    y = -quat[2]
    z = -quat[3]

    theta = np.empty((3, 3))

    # front direction
    theta[0, 0] = 1.0 - 2.0 * s * (y * y + z * z)
    theta[1, 0] = 2.0 * s * (x * y + z * w)
    theta[2, 0] = 2.0 * s * (x * z - y * w)

    # left direction
    theta[0, 1] = 2.0 * s * (x * y - z * w)
    theta[1, 1] = 1.0 - 2.0 * s * (x * x + z * z)
    theta[2, 1] = 2.0 * s * (y * z + x * w)

    # up direction
    theta[0, 2] = 2.0 * s * (x * z + y * w)
    theta[1, 2] = 2.0 * s * (y * z - x * w)
    theta[2, 2] = 1.0 - 2.0 * s * (x * x + y * y)

    return theta



from rlgym.communication import Message
from rlgym.communication import communication_exception_handler

import win32file
import win32pipe
from multiprocessing.pool import ThreadPool

class CommunicationHandler(object):
    RLGYM_GLOBAL_PIPE_ID = "RLGYM_GLOBAL_COMM_PIPE"
    RLGYM_GLOBAL_PIPE_NAME = r"\\.\pipe\RLGYM_GLOBAL_COMM_PIPE"
    RLGYM_DEFAULT_PIPE_SIZE = 4096

    def __init__(self):
        self._current_pipe_name = CommunicationHandler.RLGYM_GLOBAL_PIPE_NAME
        self._pipe = None
        self._connected = False

    def receive_message(self, header=None, num_attempts=100):
        #TODO: deal with discarded messages while waiting for a specific header
        if not self.is_connected():
            print("RLGYM ATTEMPTED TO RECEIVE MESSAGE WITH NO CONNECTION")
            return communication_exception_handler.BROKEN_PIPE_ERROR

        m = Message()
        received_message = Message()
        exception_code = None
        for i in range(num_attempts):
            try:
                code, msg_bytes = win32file.ReadFile(self._pipe, CommunicationHandler.RLGYM_DEFAULT_PIPE_SIZE)

            #This is the pywintypes.error object type.
            except BaseException as e:
                exception_code = communication_exception_handler.handle_exception(e)
                break

            msg_str = bytes.decode(msg_bytes)
            m.deserialize(msg_str)

            #Only deserialize valid messages.
            if header is None or header == m.header:
                received_message.deserialize(msg_str)
                # Peek the next message in the pipe to see if we've reached the end of new messages.
                data = win32pipe.PeekNamedPipe(self._pipe, CommunicationHandler.RLGYM_DEFAULT_PIPE_SIZE)
                if data[0] == b'':
                    break

        #TODO: make sure users of this object deal with the null message response
        return received_message, exception_code

    def send_message(self, message=None, header=None, body=None):
        if not self.is_connected():
            print("RLGYM ATTEMPTED TO SEND MESSAGE WITH NO CONNECTION")
            return communication_exception_handler.BROKEN_PIPE_ERROR

        if message is None:
            if header is None:
                header = Message.RLGYM_NULL_MESSAGE_HEADER

            if body is None:
                body = Message.RLGYM_NULL_MESSAGE_BODY

            message = Message(header=header, body=body)

        serialized = message.serialize()
        exception_code = None
        try:
            win32file.WriteFile(self._pipe, str.encode(serialized))

        except BaseException as e:
            exception_code = communication_exception_handler.handle_exception(e)

        return exception_code

    def open_pipe(self, pipe_name=None, num_allowed_instances=1):
        if pipe_name is None:
            pipe_name = CommunicationHandler.RLGYM_GLOBAL_PIPE_NAME

        if self.is_connected():
            self.close_pipe()

        self._connected = False

        pool = ThreadPool(processes=1)
        pool.apply_async(CommunicationHandler.handle_diemwin_potential, args=[self.is_connected])

        #win32pipe.PIPE_UNLIMITED_INSTANCES
        self._pipe = win32pipe.CreateNamedPipe(pipe_name,
                                               win32pipe.PIPE_ACCESS_DUPLEX | win32file.FILE_FLAG_OVERLAPPED,

                                               win32pipe.PIPE_TYPE_MESSAGE |
                                               win32pipe.PIPE_READMODE_MESSAGE |
                                               win32pipe.PIPE_WAIT,

                                               num_allowed_instances,

                                               CommunicationHandler.RLGYM_DEFAULT_PIPE_SIZE,
                                               CommunicationHandler.RLGYM_DEFAULT_PIPE_SIZE, 0, None)

        win32pipe.ConnectNamedPipe(self._pipe)

        self._current_pipe_name = pipe_name
        self._connected = True

        pool.terminate()
        pool.join()

    def close_pipe(self):
        self._connected = False
        win32file.CloseHandle(self._pipe)

    def is_connected(self):
        return self._connected

    @staticmethod
    def format_pipe_id(pipe_id):
        return r"\\.\pipe\{}".format(pipe_id)

    @staticmethod
    def handle_diemwin_potential(connected):
        import time
        import pywinauto

        # Windows Direct Input Emulator (DIEmWin) spawns a window sometimes, so we delete it here.
        while not connected():
            try:
                app = pywinauto.Application().connect(title='DIEmWin', visible_only=True)
                if app.is_process_running():
                    app.window(title='DIEmWin', visible_only=True).close()
                    print("DIEmWin detector successfully closed window")
                    return
            except:
                time.sleep(2)



from abc import ABC, abstractmethod
from rlgym.utils.gamestates import PhysicsObject, PlayerData, GameState
import numpy as np
from typing import Any


class ObsBuilder(ABC):
    @staticmethod
    def get_random_physics_state():
        means = [-7.55690032e+00, -7.67272187e+02, 3.99790479e+01, 4.33080865e-01, 3.29746054e-03, 9.68255500e-03,
                 4.10564683e-01, 1.64972058e+01, -9.52952112e+00, 1.21217917e-01, 6.23036234e-03, -1.84941779e-03,
                 -1.82993569e-02, np.pi, np.pi, np.pi]

        stds = [2.22632843e+03, 3.31466701e+03, 7.04738414e+01, 5.49264807e-01, 1.37360694e-01, 1.42669913e-01,
                5.50325801e-01, 6.42938410e+02, 7.39945204e+02, 1.34840452e+02, 7.39983198e-01, 7.07553505e-01,
                1.78035710e+00, np.pi, np.pi, np.pi]

        mu = np.asarray(means)
        sigma = np.asarray(stds)

        obs = (np.random.randn(len(means)) * sigma + mu).tolist()
        obj = PhysicsObject()
        obj.position = obs[:3]
        obj.quaternion = obs[3:7]
        obj.linear_velocity = obs[7:10]
        obj.angular_velocity = obs[10:13]
        obj._euler_angles = obs[13:]  # TODO remove this
        return obj

    def __init__(self):
        pass

    # This method is optional
    @abstractmethod
    def reset(self, optional_data=None):
        raise NotImplementedError

    # This method is optional
    @abstractmethod
    def build_obs(self, state: GameState, optional_data=None) -> np.ndarray:
        raise NotImplementedError

    # This is the function that rlgym calls to make the obs for each agent
    # A List of whatever you return here will be the value of "obs" returned from gym.step
    @abstractmethod
    def build_obs_for_player(self, player: PlayerData, state: GameState, previous_action: np.ndarray, optional_data=None) -> Any:
        raise NotImplementedError



from rlgym.envs.environment import Environment
from rlgym.utils.gamestates import GameState, PhysicsObject
from rlgym.utils import common_values
import gym.spaces
import numpy as np
from typing import List, Union


class Match(Environment):
    def __init__(self,
                 team_size=1,
                 tick_skip=8,
                 game_speed=100,
                 spawn_opponents=False,
                 random_resets=False,
                 self_play=False,
                 reward_function=None,
                 terminal_conditions=None,
                 obs_builder=None):
        super().__init__()

        self._game_speed = game_speed
        self._team_size = team_size
        self._self_play = self_play
        self._spawn_opponents = spawn_opponents
        self._tick_skip = tick_skip
        self._random_resets = random_resets
        self._reward_fn = reward_function
        self._terminal_conditions = terminal_conditions
        self._obs_builder = obs_builder

        if self._reward_fn is None:
            from rlgym.utils.reward_functions import ShootBallReward
            self._reward_fn = ShootBallReward()

        if obs_builder is None:
            from rlgym.utils.obs_builders import RhobotObs
            self._obs_builder = RhobotObs()

        if terminal_conditions is None:
            from rlgym.utils.terminal_conditions import common_conditions
            ep_len_minutes = 20 / 60
            ticks_per_sec = 120
            ticks_per_min = ticks_per_sec * 60
            max_ticks = int(round(ep_len_minutes * ticks_per_min / self._tick_skip))
            self._terminal_conditions = [common_conditions.TimeoutCondition(max_ticks),
                                         common_conditions.GoalScoredCondition()]

        elif type(terminal_conditions) not in (tuple, list):
            self._terminal_conditions = [terminal_conditions, ]

        self.agents = self._team_size * 2 if self._self_play else self._team_size
        self.observation_space = gym.spaces.Box(-np.inf, np.inf, shape=(self._obs_builder.obs_size,))
        self.action_space = gym.spaces.Box(-1, 1, shape=(8,))

        self._prev_actions = np.zeros((self.agents, self.action_space.shape[0]), dtype=float)

        self._spectator_ids = [i + 1 for i in range(self._team_size)]
        if self._self_play:
            for i in range(self._team_size):
                self._spectator_ids.append(5 + i)

        self.last_touch = None

    def episode_reset(self):
        self._prev_actions.fill(0)
        for condition in self._terminal_conditions:
            condition.reset()
        self._reward_fn.reset()
        self._obs_builder.reset()
        self.last_touch = None

    def build_observations(self, state) -> List:
        observations = []

        for i in range(len(state.players)):
            player = state.players[i]

            if not self._self_play and player.team_num == common_values.ORANGE_TEAM:
                continue
            else:
                obs = self._obs_builder.build_obs_for_player(player, state, self._prev_actions[i])

            observations.append(obs)

        if state.last_touch is None:
            state.last_touch = self.last_touch
        else:
            self.last_touch = state.last_touch
        return observations

    def get_rewards(self, state) -> List:
        rewards = []
        for player in state.players:
            if player.team_num == common_values.ORANGE_TEAM and not self._self_play:
                continue

            done = False
            for condition in self._terminal_conditions:
                if condition.look_ahead(state):
                    done = True
                    break

            if done:
                reward = self._reward_fn.get_final_reward(player, state)
            else:
                reward = self._reward_fn.get_reward(player, state)

            rewards.append(reward)
        return rewards

    def is_done(self, state):
        for condition in self._terminal_conditions:
            if condition.is_terminal(state):
                return True
        return False

    def parse_state(self, state_str: str) -> GameState:
        state = GameState(state_str)
        return state

    def format_actions(self, actions: Union[np.ndarray, List[np.ndarray], List[float]]):
        if type(actions) != np.ndarray:
            actions = np.asarray(actions)

        n = len(actions)
        if n != self.agents:
            actions = actions.reshape((1, n))

        self._prev_actions[:] = actions[:]

        action_str = None
        for i in range(len(actions)):
            act_arr = [str(self._spectator_ids[i])]
            for x in actions[i]:
                act_arr.append(str(x))
            act_str = ' '.join(act_arr)

            if action_str is None:
                action_str = act_str
            else:
                action_str = "{} {}".format(action_str, act_str)

        return action_str

    def get_config(self):
        return '{} {} {} {} {} {}'.format(self._team_size,
                                          1 if self._self_play else 0,
                                          1 if self._spawn_opponents else 0,
                                          1 if self._random_resets else 0,
                                          self._tick_skip,
                                          self._game_speed
                                          )


from setuptools import setup, find_packages
from setuptools.command.install import install


__version__ = None  # This will get replaced when reading version.py
exec(open('rlgym/version.py').read())


with open('README.md', 'r') as readme_file:
    long_description = readme_file.read()


class CustomInstall(install):
    def run(self):
        install.run(self)
        self.install_plugin()

    def install_plugin(self):
        print('Installing plugin')
        try:
            import os
            import sys
            import winreg
            import shutil
            import pkg_resources

            # Get bakkesmod folder
            bm_path = winreg.QueryValueEx(
                        winreg.OpenKey(winreg.HKEY_CURRENT_USER, 'Software\\BakkesMod\\AppPath'), 'BakkesModPath')[0]

            # Install RLGym plugin
            dll_path = pkg_resources.resource_filename('rlgym', 'plugin/RLGym.dll')
            print('dll_path', dll_path)
            shutil.copy2(dll_path, os.path.join(bm_path, 'plugins'))

            # Enable RLGym plugin
            bm_config_path = os.path.join(bm_path, 'cfg/plugins.cfg')
            print('bm_config_path', bm_config_path)
            with open(bm_config_path, 'r') as bm_config:
                content = bm_config.read()
                enabled = 'plugin load rlgym' in content

            print('enabled', enabled)
            if not enabled:
                with open(bm_config_path, 'a') as bm_config:
                    bm_config.write('plugin load rlgym\n')
        except:
            raise RuntimeError('Plugin installation failed')


setup(
    name='rlgym',
    packages=find_packages(),
    version=__version__,
    description='A python API that can be used to treat the game Rocket League as an Openai Gym-like environment for '
                'Reinforcement Learning projects.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Lucas Emery and Matthew Allen',
    url='https://github.com/lucas-emery/rocket-league-gym',
    install_requires=[
        'gym>=0.17',
        'numpy>=1.19',
        'pywin32==228',
        'pywinauto==0.6.8',
        'psutil>=5.8',
    ],
    python_requires='>=3.7',
    cmdclass={'install': CustomInstall},
    license='Apache 2.0',
    license_file='LICENSE',
    keywords=['rocket-league', 'gym', 'reinforcement-learning'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        "Operating System :: Microsoft :: Windows",
    ],
    package_data={
        'rlgym': [
            'plugin/*'
        ]
    }
)


from rlgym.envs import match_configs as conf
from rlgym.envs import Match


def build_match(match_id, **kwargs):
    m = match_id.strip().lower()

    if "self" in m:
        kwargs["self_play"] = True

    preconfigured_match = None
    if "duel" in m:
        preconfigured_match = conf.basic_duel_match
    elif "double" in m:
        preconfigured_match = conf.basic_doubles_match
    elif "standard" in m:
        preconfigured_match = conf.basic_standard_match

    if preconfigured_match is not None:
        match = preconfigured_match(**kwargs)
    else:
        game_speed, tick_skip, spawn_opponents, random_resets, self_play, team_size, terminal_conditions, reward_fn, obs_builder \
            = conf.get_default_params(**kwargs)

        match = Match(team_size=team_size,
                      tick_skip=tick_skip,
                      game_speed=game_speed,
                      spawn_opponents=spawn_opponents,
                      random_resets=random_resets,
                      self_play=self_play,
                      reward_function=reward_fn,
                      terminal_conditions=terminal_conditions,
                      obs_builder=obs_builder)

    return match



import numpy as np
from typing import List, Optional
from rlgym.utils.gamestates import PlayerData, PhysicsObject


class GameState(object):
    BALL_STATE_LENGTH = 18
    PLAYER_INFO_LENGTH = 38
    PLAYER_CAR_STATE_LENGTH = 13
    PLAYER_TERTIARY_INFO_LENGTH = 10

    def __init__(self, state_str: str):
        self.game_type: int = 0
        self.blue_score: int = -1
        self.orange_score: int = -1
        self.last_touch: Optional[int] = None

        self.players: List[PlayerData] = []

        self.ball: PhysicsObject = PhysicsObject()
        self.inv_ball: PhysicsObject = PhysicsObject()

        self.decode(state_str)

    def decode(self, state_str: str):
        assert type(state_str) == str, "UNABLE TO DECODE STATE OF TYPE {}".format(type(state_str))
        self._decode(state_str)

    @staticmethod
    def _decode_state_str(state_str: str) -> np.ndarray:
        delimiter = ' '
        split_state = state_str.strip(delimiter).split(delimiter)
        return np.array(split_state, dtype=np.float32)

    def _decode(self, state_str: str):
        p_len = GameState.PLAYER_INFO_LENGTH
        b_len = GameState.BALL_STATE_LENGTH
        start = 3

        state_vals = GameState._decode_state_str(state_str)
        num_ball_packets = 1
        # The state will contain the ball, the mirrored ball, every player, every player mirrored, the score for both teams, and the number of ticks since the last packet was sent.
        num_player_packets = int((len(state_vals) - num_ball_packets * b_len - start) / p_len)
        #print(len(state_vals), " | ", num_ball_packets, " | ", num_player_packets)

        #print(state_str)

        ticks = int(state_vals[0])
        # print(ticks)
        self.blue_score = int(state_vals[1])
        self.orange_score = int(state_vals[2])

        ball_data = state_vals[start:start + b_len]
        # print("BALL:",ball_data)
        self.ball.decode_ball_data(ball_data)
        start += b_len // 2

        inv_ball_data = state_vals[start:start + b_len]
        # print("INV_BALL:",inv_ball_data)
        self.inv_ball.decode_ball_data(inv_ball_data)
        start += b_len // 2

        for i in range(num_player_packets):
            player = self.decode_player(state_vals[start:start + p_len])
            self.players.append(player)
            start += p_len

            if player.ball_touched:
                self.last_touch = player.car_id

        #print("State decoded!")
        #print(self)


    def decode_player(self, full_player_data):
        #print("DECODING PLAYER ",full_player_data)
        player_data = PlayerData()
        c_len = GameState.PLAYER_CAR_STATE_LENGTH
        t_len = GameState.PLAYER_TERTIARY_INFO_LENGTH

        start = 2

        car_data = full_player_data[start:start + c_len]
        player_data.car_data.decode_car_data(car_data)
        start += c_len

        inv_state_data = full_player_data[start:start + c_len]
        player_data.inverted_car_data.decode_car_data(inv_state_data)
        start += c_len

        tertiary_data = full_player_data[start:start + t_len]

        player_data.match_goals = int(tertiary_data[0])
        player_data.match_saves = int(tertiary_data[1])
        player_data.match_shots = int(tertiary_data[2])
        player_data.match_demolishes = int(tertiary_data[3])
        player_data.boost_pickups = int(tertiary_data[4])
        player_data.is_alive = True if tertiary_data[5] > 0 else False
        player_data.on_ground = True if tertiary_data[6] > 0 else False
        player_data.ball_touched = True if tertiary_data[7] > 0 else False
        player_data.has_flip = True if tertiary_data[8] > 0 else False
        player_data.boost_amount = float(tertiary_data[9])
        player_data.car_id = int(full_player_data[0])
        player_data.team_num = int(full_player_data[1])

        return player_data

    def __str__(self):
        output = "{}GAME STATE OBJECT{}\n" \
                 "Game Type: {}\n" \
                 "Orange Score: {}\n" \
                 "Blue Score: {}\n" \
                 "PLAYERS: {}\n" \
                 "BALL: {}\n" \
                 "INV_BALL: {}\n" \
                 "".format("*" * 8, "*" * 8,
                           self.game_type,
                           self.orange_score,
                           self.blue_score,
                           self.players,
                           self.ball,
                           self.inv_ball)

        return output



import platform
from dataclasses import dataclass
import subprocess
import webbrowser
from pathlib import Path
from typing import Optional
from rlgym.gamelaunch.epic_launch import launch_with_epic_simple, launch_with_epic_login_trick
import os

# Copied from https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/setup_manager.py

class ROCKET_LEAGUE_PROCESS_INFO:
    GAMEID = 252950
    PROGRAM_NAME = 'RocketLeague.exe'
    PROGRAM = 'RocketLeague.exe'
    REQUIRED_ARGS = {r'-pipe', r'[0-9]+'}

    @staticmethod
    def get_ideal_args(pipe_id):
        # We are specifying RLBot_PacketSendRate=240, which will override people's TARLBot.ini settings.
        # We believe there is no downside to 240. See https://github.com/RLBot/RLBot/wiki/Tick-Rate
        return ['-pipe', f'{pipe_id}', '-nomovie']

@dataclass
class RocketLeagueLauncherPreference:
    STEAM = 'steam'
    EPIC = 'epic'
    preferred_launcher: str
    use_login_tricks: bool

# By default, we will attempt Epic with no login tricks, then fall back to Steam.
DEFAULT_LAUNCHER_PREFERENCE = RocketLeagueLauncherPreference(RocketLeagueLauncherPreference.EPIC, False)


def launch_rocket_league(pipe_id, path_to_rl=None, use_injector=False, launcher_preference: RocketLeagueLauncherPreference = DEFAULT_LAUNCHER_PREFERENCE):
    """
    Launches Rocket League but does not connect to it.
    """
    ideal_args = ROCKET_LEAGUE_PROCESS_INFO.get_ideal_args(pipe_id)

    if path_to_rl:
        if os.path.isfile(path_to_rl):
            game_process = subprocess.Popen([path_to_rl] + ideal_args)
            if use_injector:
                print("Executing injector...")
                # TODO add RL exe name as parameter for injector
                injector_command = os.path.join(os.path.join(path_to_rl, os.pardir), "RLMultiInjector.exe")
                subprocess.Popen(injector_command)
            return game_process
        else:
            print("path_to_rl doesn't point to RocketLeague.exe")

    if launcher_preference.preferred_launcher == RocketLeagueLauncherPreference.EPIC:
        if launcher_preference.use_login_tricks:
            if launch_with_epic_login_trick(ideal_args):
                return
            else:
                print("Epic login trick seems to have failed, falling back to simple Epic launch.")
        # Fall back to simple if the tricks failed or we opted out of tricks.
        if launch_with_epic_simple(ideal_args):
            return

    # Try launch via Steam.
    steam_exe_path = try_get_steam_executable_path()
    if steam_exe_path:  # Note: This Python 3.8 feature would be useful here https://www.python.org/dev/peps/pep-0572/#abstract
        exe_and_args = [
                           str(steam_exe_path),
                           '-applaunch',
                           str(ROCKET_LEAGUE_PROCESS_INFO.GAMEID),
                       ] + ideal_args
        # print(f'Launching Rocket League with: {exe_and_args}')
        _ = subprocess.Popen(exe_and_args)  # This is deliberately an orphan process.
        return

    print(f'Launching Rocket League using Steam-only fall-back launch method with args: {ideal_args}')
    print("You should see a confirmation pop-up, if you don't see it then click on Steam! "
                     'https://gfycat.com/AngryQuickFinnishspitz')
    args_string = '%20'.join(ideal_args)

    # Try launch via terminal (Linux)
    if platform.system() == 'Linux':
        linux_args = [
            'steam',
            f'steam://rungameid/{ROCKET_LEAGUE_PROCESS_INFO.GAMEID}//{args_string}'
        ]

        try:
            _ = subprocess.Popen(linux_args)
            return
        except OSError:
            print('Could not launch Steam executable on Linux.')

    try:
        print("Launching rocket league via steam browser URL as a last resort...")
        webbrowser.open(f'steam://rungameid/{ROCKET_LEAGUE_PROCESS_INFO.GAMEID}//{args_string}')
    except webbrowser.Error:
        print(f'Unable to launch Rocket League. Please launch Rocket League manually using the -pipe {pipe_id} option to continue.')

def try_get_steam_executable_path() -> Optional[Path]:
    """
    Tries to find the path of the Steam executable.
    Has platform specific code.
    """

    try:
        from winreg import OpenKey, HKEY_CURRENT_USER, ConnectRegistry, QueryValueEx, REG_SZ
    except ImportError as e:
        return  # TODO: Linux support.

    try:
        key = OpenKey(ConnectRegistry(None, HKEY_CURRENT_USER), r'Software\Valve\Steam')
        val, val_type = QueryValueEx(key, 'SteamExe')
    except FileNotFoundError:
        return
    if val_type != REG_SZ:
        return
    return Path(val)